import epics
import json
import time
import sys
import numpy
import pandas
import datetime
import requests
from threading import Thread


def p2t(p):
    # This formula uses the Table 3 of ITS-90 from NIST paper: https://www.nist.gov/system/files/documents/pml/div685/grp01/ITS-90_metrologia.pdf
    if not numpy.isreal(p):  # If pressure is not a number
        return "N/A"
    if p < 1.14734339634282:  # This formula is valid only for Temperature > 1.25 K
        return "N/A"
    if p > 1960.16532874851:  # 'This formula is valid only for Temperature < 5 K'
        return "N/A"
    if p <= 50.4181:
        P = p * 100  # From mbar to Pascal
        A = [
            1.392408,
            0.527153,
            0.166756,
            0.050988,
            0.026514,
            0.001975,
            -0.017976,
            0.005409,
            0.013259,
            0.000000,
        ]
        B = 5.6
        C = 2.9
        t = A[0]
        for i in range(1, 10):
            t += A[i] * ((numpy.log(P) - B) / C) ** i
    else:
        P = p * 100  # From mbar to Pascal
        A = [
            3.146631,
            1.357655,
            0.413923,
            0.091159,
            0.016349,
            0.001826,
            -0.004325,
            -0.004973,
            0.000000,
            0.000000,
        ]
        B = 10.3
        C = 1.9
        t = A[0]
        for i in range(1, 10):
            t += A[i] * ((numpy.log(P) - B) / C) ** i
    return t


def get_from_archiver(pv, start_date, end_date):
    try:
        url = "http://archiver-01.tn.esss.lu.se:17668/retrieval/data/getData.json?pv=ncount({})&from={}&to={}".format(
            pv, start_date, end_date
        )
        entries = int(requests.get(url).json()[0]["data"][0]["val"])
        url = "http://archiver-01.tn.esss.lu.se:17668/retrieval/data/getData.json?pv=nth_{}({})&from={}&to={}".format(
            int(entries // 100), pv, start_date, end_date
        )
        r = requests.get(url).json()[0]["data"]
        raw_data = pandas.DataFrame.from_dict(r)
        return list(
            zip(
                (raw_data["secs"] * 1000).astype("int")
                + (raw_data["nanos"] / 1e6).astype("int"),
                round(raw_data["val"], 3),
            )
        )
    except Exception:
        return []


class ts2_helevpresScreen(Thread):
    def __init__(self, stop_signal):
        Thread.__init__(self)
        self.stop_signal = stop_signal

    def run(self):
        pvs = [
            "TS2-010CRM:Cryo-PT-003:MeasValue",
            "TS2-010CRM:Cryo-PT-004:MeasValue",
            "TS2-010CRM:Cryo-LT-001:MeasValue",
            "TS2-010CRM:Cryo-LT-002:MeasValue",
        ]
        intervals = {"2w": 24 * 14, "1w": 24 * 7, "2d": 24 * 2, "1d": 24, "live": 1}

        epics_dict = {}
        for pv in pvs:
            epics_dict[pv] = epics.PV(pv, auto_monitor=True)

        json_dict = {}
        counter = 0
        while not self.stop_signal.isSet():
            try:
                for pv in epics_dict:
                    if pv == "TS2-010CRM:Cryo-PT-003:MeasValue":
                        json_dict[pv] = {}
                        if epics_dict[pv].connected:
                            json_dict[pv]["units"] = epics_dict[pv].units
                            json_dict[pv]["timestamp"] = epics_dict[pv].timestamp
                            json_dict[pv]["temp"] = "N/A"
                            value = epics_dict[pv].value
                            if type(value) == float:
                                json_dict[pv]["value"] = round(value, 3)
                                json_dict[pv]["temp"] = round(
                                    p2t(json_dict[pv]["value"]), 3
                                )
                            else:
                                json_dict[pv]["value"] = value
                    if pv == "TS2-010CRM:Cryo-PT-004:MeasValue":
                        json_dict[pv] = {}
                        if epics_dict[pv].connected:
                            json_dict[pv]["units"] = epics_dict[pv].units
                            json_dict[pv]["timestamp"] = epics_dict[pv].timestamp
                            json_dict[pv]["temp"] = "N/A"
                            value = epics_dict[pv].value
                            if type(value) == float:
                                json_dict[pv]["value"] = round(value, 3)
                                if value <= 65:
                                    json_dict[pv]["temp"] = round(
                                        p2t(json_dict[pv]["value"]), 3
                                    )
                            else:
                                json_dict[pv]["value"] = value
                    if (
                        pv == "TS2-010CRM:Cryo-LT-001:MeasValue"
                        or pv == "TS2-010CRM:Cryo-LT-002:MeasValue"
                    ):
                        if counter % 600 == 0:
                            json_dict[pv] = {}
                            for interval in intervals:
                                start_date = (
                                    datetime.datetime.now()
                                    - datetime.timedelta(hours=intervals[interval])
                                ).isoformat() + "Z"
                                end_date = datetime.datetime.now().isoformat() + "Z"
                                json_dict[pv][interval] = get_from_archiver(
                                    pv, start_date, end_date
                                )
                            counter = 0
                        else:
                            if epics_dict[pv].connected:
                                value = epics_dict[pv].value
                                timestamp = int(epics_dict[pv].timestamp * 1000)
                                if (
                                    type(value) == float
                                    and timestamp != json_dict[pv]["live"][-1][0]
                                ):
                                    json_dict[pv]["live"] = json_dict[pv]["live"][1:]
                                    json_dict[pv]["live"].append(
                                        (timestamp, round(value, 3))
                                    )

                tmp_json = json.dumps(json_dict)
                tmp_json = tmp_json.replace("NaN", "0")
                tmp_json = tmp_json.replace("Infinity", "0")
                with open("/var/www/data/ts2_helevpres.json", "w") as datafile:
                    datafile.write(tmp_json)
            except Exception:
                print(sys.exc_info())
                time.sleep(1)
            time.sleep(0.5)
            counter += 1

        for pv in epics_dict:
            epics_dict[pv].clear_auto_monitor()
            epics_dict[pv].disconnect()
